/*
 *  $Id: cdpeek.c,v 1.2 2004/05/25 18:15:55 mike Exp $
 *
 *  cdppeek
 *  CDP information discovery tool
 *  Released as part of the MXFP Layer 2 Toolkit
 *  http://www.packetfactory.net/MXFP
 *
 *  Copyright (c) 2004 Mike D. Schiffman  <mike@infonexus.com>
 *  Copyright (c) 2004 Jeremy Rauch       <jrauch@cadre.org>
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <pcap.h>
#include <libnet.h>

#define  SNAPLEN  8192
#define  PROMISC  1
#define  BIT(n)   (1 << n)

#define CDP_FILTER "ether dst 1:0:c:cc:cc:cc"

void hexdump(char *addr, int length);
char *type_to_string(int type);
void cdp_decode(const u_char *packet, const struct pcap_pkthdr *pc_hdr);
int apply_pcap_filter(char *device, char *filter, pcap_t *p, char *errbuf);
static void packet_process(u_char *pcap_data, const struct
pcap_pkthdr *packet_header, const u_char *packet);

int
main(int argc, char **argv)
{
    pcap_t *p;
    char pcap_error[PCAP_ERRBUF_SIZE];
    int return_value;
    time_t t;

    if (argc != 2)
    {
        fprintf(stderr, "usage: %s interface\n", argv[0]);
        return (EXIT_FAILURE);
    }

    /* initialize pcap */
    p = pcap_open_live (argv[1], SNAPLEN, PROMISC, 1000, pcap_error);
    if (p == NULL) 
    {
        fprintf (stderr, "pcap_open_live(): %s\n", pcap_error);
        return (EXIT_FAILURE);
    }
  
    if (apply_pcap_filter(argv[1], CDP_FILTER, p, pcap_error) == -1)
    {
        fprintf(stderr, "%s\n", pcap_error);
        pcap_close(p);
        return (EXIT_FAILURE);
    }

    t = time((time_t *)NULL);
    fprintf(stderr, "cdppeek running: %s", ctime(&t));

    return_value = pcap_loop(p, -1, packet_process, NULL);
    if (!return_value)
    {
        fprintf (stderr, " Error : pcap_dispatch (%d)\n", return_value);
        return (EXIT_FAILURE);
    }

    return (EXIT_SUCCESS);
} 

char *
type_to_string(int type)
{
    switch(type)
    {
        case 0x0001:
            return ("device id");
        case 0x0002:
            return ("address");
        case 0x0003:
            return ("port id");
        case 0x0004:
            return ("capabils");
        case 0x0005:
            return ("version");
        case 0x0006:
            return ("platform");
        case 0x0007:
            return ("ip prefix");
        default:
            return ("oops we dont support v2 yet");
    }
    return (NULL);

}

void
cdp_decode(const u_char *packet, const struct pcap_pkthdr *pc_hdr)
{
    int i;
    static int packet_count;
    u_char *s, *e;
    short type, length = 0;
    char buf[1024];

    packet_count++;

    /* step over ethernet header and SNAP header, the start of the CDP frame */
    s = (u_char*) packet + LIBNET_802_3_H + LIBNET_802_2SNAP_H;

    /* point to the end of the frame */
    e = (u_char*) packet + pc_hdr->len;

    /* print packet count */
    fprintf(stderr, "[frame %03d] ", packet_count);

    /* print the source MAC address */
    for (i = 0; i < 6; i++)
    {
        fprintf(stderr, "%02x", packet[i + 6]);
        if (i != 5)
        {
            fprintf(stderr, ":");
        }
    }

    /* print toplevel header information */
    fprintf(stderr, " %d bytes\tCDP version: %02d\tTTL: %02d ",
            pc_hdr->len, *s, *(s + 1));

    /* skip over the base header to the first TLV */
    s += 4;

    /* we're pointing to the first TLV */
    fprintf(stderr, "\n");
    while (s < e)
    {
        /* 
         * On some platforms, when receiving 802.3/802.2 packets the trailer
         * shows up, which can make a mess so we handle it here.
         */
        if ((e - s) == 4)
        {
          fprintf(stderr, "\n");
          return;
        }

        /* pull in the type */
        memcpy(&type, s, 2);
        type = htons(type);
        s += 2;

        /* pull in the length */
        memcpy(&length, s, 2);
        length = htons(length);
        s += 2;

        /* handle bad frame */
        if ((s + length - 4) > e)
        {
            fprintf(stderr, "Malformed packet %p %p\n", s + length - 4, e);
            return;
        }

        /* dump the type and length */
        fprintf(stderr, "type: %s\t\tlength: %d\tvalue: ",
                type_to_string(type), length);

        switch (type)
        {
            case 0x001:
            case 0x003:
            case 0x005:
            case 0x006:
                /* all ASCII types we printf */
                memset(buf, 0, 1024);
                memcpy(buf, s, length - 4); 
                fprintf(stderr, "%s\n", buf);
                break;
            case 0x02:
            case 0x04:
            case 0x07:
                /* all non ASCII types we hexdump */
                hexdump(s, length - 4);
                break;
            default:
                /* unknown type */
                fprintf(stderr, "\n");
                return;
        }
        s += length - 4;
    }
    fprintf(stderr, "\n");
}

int
apply_pcap_filter(char *device, char *filter, pcap_t *p, char *errbuf)
{
    char err[PCAP_ERRBUF_SIZE];
    struct bpf_program filter_code;
    bpf_u_int32 local_net, netmask;

    /* get the subnet mask of the interface */
    if (pcap_lookupnet(device, &local_net, &netmask, err) == -1)
    {
        sprintf(errbuf, "pcap_lookupnet(): %s", err);
        return (-1);
    }

    /* compile the BPF filter code */
    if (pcap_compile(p, &filter_code, CDP_FILTER, 1, netmask) == -1)
    {
        sprintf(errbuf, pcap_geterr(p));
        return (-1);
    }

    /* apply the filter to the interface */
    if (pcap_setfilter(p, &filter_code) == -1)
    {
        sprintf(errbuf, pcap_geterr(p));
        return (-1);
    }
    return (1);
}

static void 
packet_process(u_char *pcap_data, const struct pcap_pkthdr *packet_header,
const u_char *packet)
{
    cdp_decode(packet, packet_header);   
}

void
hexdump(char *addr, int length)
{
    int p;
    char asc[9];
    fprintf(stderr, "\n\t\t\t");
    asc[8] = 0;

    for (p = 0; p < length; p++)
    {
        fprintf(stderr, "%2.2x ", (unsigned char)* (addr + p));

        if (isalnum((unsigned char)* (addr + p)))
        {
            asc[p % 8] = (unsigned char)*(addr + p);
        }
        else
        {
            asc[p % 8] = '.';
        }
        if (p % 8 == 7)
        {
            fprintf(stderr, "\t\t%s",asc);
            memset(asc, 0, 9);
            fprintf(stderr, "\n\t\t\t");
        }
    }
  
    for (p = 0; p < length % 8; p++)
    {
        fprintf(stderr, "   "); 
    }
    fprintf(stderr, "\t\t%s", asc);
    fprintf(stderr, "\n");
}

/* EOF */
